# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Graph(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Graph()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsGraph(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Graph
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Graph
    def Node(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.Node import Node
            obj = Node()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Graph
    def NodeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Graph
    def NodeIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Graph
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Graph
    def Initializer(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Graph
    def InitializerLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Graph
    def InitializerIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Graph
    def DocString(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Graph
    def Input(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.ValueInfo import ValueInfo
            obj = ValueInfo()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Graph
    def InputLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Graph
    def InputIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Graph
    def Output(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.ValueInfo import ValueInfo
            obj = ValueInfo()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Graph
    def OutputLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Graph
    def OutputIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # Graph
    def ValueInfo(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.ValueInfo import ValueInfo
            obj = ValueInfo()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Graph
    def ValueInfoLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Graph
    def ValueInfoIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0

    # Graph
    def QuantizationAnnotation(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.TensorAnnotation import TensorAnnotation
            obj = TensorAnnotation()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Graph
    def QuantizationAnnotationLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Graph
    def QuantizationAnnotationIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0

    # Graph
    def TestInputsValue(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Graph
    def TestInputsValueLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Graph
    def TestInputsValueIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        return o == 0

    # Graph
    def TestOutputsValue(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Graph
    def TestOutputsValueLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Graph
    def TestOutputsValueIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        return o == 0

def GraphStart(builder):
    builder.StartObject(10)

def Start(builder):
    GraphStart(builder)

def GraphAddNode(builder, node):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(node), 0)

def AddNode(builder, node):
    GraphAddNode(builder, node)

def GraphStartNodeVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartNodeVector(builder, numElems):
    return GraphStartNodeVector(builder, numElems)

def GraphAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)

def AddName(builder, name):
    GraphAddName(builder, name)

def GraphAddInitializer(builder, initializer):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(initializer), 0)

def AddInitializer(builder, initializer):
    GraphAddInitializer(builder, initializer)

def GraphStartInitializerVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartInitializerVector(builder, numElems):
    return GraphStartInitializerVector(builder, numElems)

def GraphAddDocString(builder, docString):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(docString), 0)

def AddDocString(builder, docString):
    GraphAddDocString(builder, docString)

def GraphAddInput(builder, input):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(input), 0)

def AddInput(builder, input):
    GraphAddInput(builder, input)

def GraphStartInputVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartInputVector(builder, numElems):
    return GraphStartInputVector(builder, numElems)

def GraphAddOutput(builder, output):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(output), 0)

def AddOutput(builder, output):
    GraphAddOutput(builder, output)

def GraphStartOutputVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartOutputVector(builder, numElems):
    return GraphStartOutputVector(builder, numElems)

def GraphAddValueInfo(builder, valueInfo):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(valueInfo), 0)

def AddValueInfo(builder, valueInfo):
    GraphAddValueInfo(builder, valueInfo)

def GraphStartValueInfoVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartValueInfoVector(builder, numElems):
    return GraphStartValueInfoVector(builder, numElems)

def GraphAddQuantizationAnnotation(builder, quantizationAnnotation):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(quantizationAnnotation), 0)

def AddQuantizationAnnotation(builder, quantizationAnnotation):
    GraphAddQuantizationAnnotation(builder, quantizationAnnotation)

def GraphStartQuantizationAnnotationVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartQuantizationAnnotationVector(builder, numElems):
    return GraphStartQuantizationAnnotationVector(builder, numElems)

def GraphAddTestInputsValue(builder, testInputsValue):
    builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(testInputsValue), 0)

def AddTestInputsValue(builder, testInputsValue):
    GraphAddTestInputsValue(builder, testInputsValue)

def GraphStartTestInputsValueVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartTestInputsValueVector(builder, numElems):
    return GraphStartTestInputsValueVector(builder, numElems)

def GraphAddTestOutputsValue(builder, testOutputsValue):
    builder.PrependUOffsetTRelativeSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(testOutputsValue), 0)

def AddTestOutputsValue(builder, testOutputsValue):
    GraphAddTestOutputsValue(builder, testOutputsValue)

def GraphStartTestOutputsValueVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartTestOutputsValueVector(builder, numElems):
    return GraphStartTestOutputsValueVector(builder, numElems)

def GraphEnd(builder):
    return builder.EndObject()

def End(builder):
    return GraphEnd(builder)

import FlatBuffers.Dl.Node
import FlatBuffers.Dl.Tensor
import FlatBuffers.Dl.TensorAnnotation
import FlatBuffers.Dl.ValueInfo
try:
    from typing import List
except:
    pass

class GraphT(object):

    # GraphT
    def __init__(self):
        self.node = None  # type: List[FlatBuffers.Dl.Node.NodeT]
        self.name = None  # type: str
        self.initializer = None  # type: List[FlatBuffers.Dl.Tensor.TensorT]
        self.docString = None  # type: str
        self.input = None  # type: List[FlatBuffers.Dl.ValueInfo.ValueInfoT]
        self.output = None  # type: List[FlatBuffers.Dl.ValueInfo.ValueInfoT]
        self.valueInfo = None  # type: List[FlatBuffers.Dl.ValueInfo.ValueInfoT]
        self.quantizationAnnotation = None  # type: List[FlatBuffers.Dl.TensorAnnotation.TensorAnnotationT]
        self.testInputsValue = None  # type: List[FlatBuffers.Dl.Tensor.TensorT]
        self.testOutputsValue = None  # type: List[FlatBuffers.Dl.Tensor.TensorT]

    @classmethod
    def InitFromBuf(cls, buf, pos):
        graph = Graph()
        graph.Init(buf, pos)
        return cls.InitFromObj(graph)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, graph):
        x = GraphT()
        x._UnPack(graph)
        return x

    # GraphT
    def _UnPack(self, graph):
        if graph is None:
            return
        if not graph.NodeIsNone():
            self.node = []
            for i in range(graph.NodeLength()):
                if graph.Node(i) is None:
                    self.node.append(None)
                else:
                    node_ = FlatBuffers.Dl.Node.NodeT.InitFromObj(graph.Node(i))
                    self.node.append(node_)
        self.name = graph.Name()
        if not graph.InitializerIsNone():
            self.initializer = []
            for i in range(graph.InitializerLength()):
                if graph.Initializer(i) is None:
                    self.initializer.append(None)
                else:
                    tensor_ = FlatBuffers.Dl.Tensor.TensorT.InitFromObj(graph.Initializer(i))
                    self.initializer.append(tensor_)
        self.docString = graph.DocString()
        if not graph.InputIsNone():
            self.input = []
            for i in range(graph.InputLength()):
                if graph.Input(i) is None:
                    self.input.append(None)
                else:
                    valueInfo_ = FlatBuffers.Dl.ValueInfo.ValueInfoT.InitFromObj(graph.Input(i))
                    self.input.append(valueInfo_)
        if not graph.OutputIsNone():
            self.output = []
            for i in range(graph.OutputLength()):
                if graph.Output(i) is None:
                    self.output.append(None)
                else:
                    valueInfo_ = FlatBuffers.Dl.ValueInfo.ValueInfoT.InitFromObj(graph.Output(i))
                    self.output.append(valueInfo_)
        if not graph.ValueInfoIsNone():
            self.valueInfo = []
            for i in range(graph.ValueInfoLength()):
                if graph.ValueInfo(i) is None:
                    self.valueInfo.append(None)
                else:
                    valueInfo_ = FlatBuffers.Dl.ValueInfo.ValueInfoT.InitFromObj(graph.ValueInfo(i))
                    self.valueInfo.append(valueInfo_)
        if not graph.QuantizationAnnotationIsNone():
            self.quantizationAnnotation = []
            for i in range(graph.QuantizationAnnotationLength()):
                if graph.QuantizationAnnotation(i) is None:
                    self.quantizationAnnotation.append(None)
                else:
                    tensorAnnotation_ = FlatBuffers.Dl.TensorAnnotation.TensorAnnotationT.InitFromObj(graph.QuantizationAnnotation(i))
                    self.quantizationAnnotation.append(tensorAnnotation_)
        if not graph.TestInputsValueIsNone():
            self.testInputsValue = []
            for i in range(graph.TestInputsValueLength()):
                if graph.TestInputsValue(i) is None:
                    self.testInputsValue.append(None)
                else:
                    tensor_ = FlatBuffers.Dl.Tensor.TensorT.InitFromObj(graph.TestInputsValue(i))
                    self.testInputsValue.append(tensor_)
        if not graph.TestOutputsValueIsNone():
            self.testOutputsValue = []
            for i in range(graph.TestOutputsValueLength()):
                if graph.TestOutputsValue(i) is None:
                    self.testOutputsValue.append(None)
                else:
                    tensor_ = FlatBuffers.Dl.Tensor.TensorT.InitFromObj(graph.TestOutputsValue(i))
                    self.testOutputsValue.append(tensor_)

    # GraphT
    def Pack(self, builder):
        if self.node is not None:
            nodelist = []
            for i in range(len(self.node)):
                nodelist.append(self.node[i].Pack(builder))
            GraphStartNodeVector(builder, len(self.node))
            for i in reversed(range(len(self.node))):
                builder.PrependUOffsetTRelative(nodelist[i])
            node = builder.EndVector()
        if self.name is not None:
            name = builder.CreateString(self.name)
        if self.initializer is not None:
            initializerlist = []
            for i in range(len(self.initializer)):
                initializerlist.append(self.initializer[i].Pack(builder))
            GraphStartInitializerVector(builder, len(self.initializer))
            for i in reversed(range(len(self.initializer))):
                builder.PrependUOffsetTRelative(initializerlist[i])
            initializer = builder.EndVector()
        if self.docString is not None:
            docString = builder.CreateString(self.docString)
        if self.input is not None:
            inputlist = []
            for i in range(len(self.input)):
                inputlist.append(self.input[i].Pack(builder))
            GraphStartInputVector(builder, len(self.input))
            for i in reversed(range(len(self.input))):
                builder.PrependUOffsetTRelative(inputlist[i])
            input = builder.EndVector()
        if self.output is not None:
            outputlist = []
            for i in range(len(self.output)):
                outputlist.append(self.output[i].Pack(builder))
            GraphStartOutputVector(builder, len(self.output))
            for i in reversed(range(len(self.output))):
                builder.PrependUOffsetTRelative(outputlist[i])
            output = builder.EndVector()
        if self.valueInfo is not None:
            valueInfolist = []
            for i in range(len(self.valueInfo)):
                valueInfolist.append(self.valueInfo[i].Pack(builder))
            GraphStartValueInfoVector(builder, len(self.valueInfo))
            for i in reversed(range(len(self.valueInfo))):
                builder.PrependUOffsetTRelative(valueInfolist[i])
            valueInfo = builder.EndVector()
        if self.quantizationAnnotation is not None:
            quantizationAnnotationlist = []
            for i in range(len(self.quantizationAnnotation)):
                quantizationAnnotationlist.append(self.quantizationAnnotation[i].Pack(builder))
            GraphStartQuantizationAnnotationVector(builder, len(self.quantizationAnnotation))
            for i in reversed(range(len(self.quantizationAnnotation))):
                builder.PrependUOffsetTRelative(quantizationAnnotationlist[i])
            quantizationAnnotation = builder.EndVector()
        if self.testInputsValue is not None:
            testInputsValuelist = []
            for i in range(len(self.testInputsValue)):
                testInputsValuelist.append(self.testInputsValue[i].Pack(builder))
            GraphStartTestInputsValueVector(builder, len(self.testInputsValue))
            for i in reversed(range(len(self.testInputsValue))):
                builder.PrependUOffsetTRelative(testInputsValuelist[i])
            testInputsValue = builder.EndVector()
        if self.testOutputsValue is not None:
            testOutputsValuelist = []
            for i in range(len(self.testOutputsValue)):
                testOutputsValuelist.append(self.testOutputsValue[i].Pack(builder))
            GraphStartTestOutputsValueVector(builder, len(self.testOutputsValue))
            for i in reversed(range(len(self.testOutputsValue))):
                builder.PrependUOffsetTRelative(testOutputsValuelist[i])
            testOutputsValue = builder.EndVector()
        GraphStart(builder)
        if self.node is not None:
            GraphAddNode(builder, node)
        if self.name is not None:
            GraphAddName(builder, name)
        if self.initializer is not None:
            GraphAddInitializer(builder, initializer)
        if self.docString is not None:
            GraphAddDocString(builder, docString)
        if self.input is not None:
            GraphAddInput(builder, input)
        if self.output is not None:
            GraphAddOutput(builder, output)
        if self.valueInfo is not None:
            GraphAddValueInfo(builder, valueInfo)
        if self.quantizationAnnotation is not None:
            GraphAddQuantizationAnnotation(builder, quantizationAnnotation)
        if self.testInputsValue is not None:
            GraphAddTestInputsValue(builder, testInputsValue)
        if self.testOutputsValue is not None:
            GraphAddTestOutputsValue(builder, testOutputsValue)
        graph = GraphEnd(builder)
        return graph



from collections import defaultdict
from typing import Callable, Dict, Iterable, List

import torch
from tqdm import tqdm

from esp_ppq.core import *
from esp_ppq.executor import BaseGraphExecutor, TorchExecutor
from esp_ppq.IR import BaseGraph, Operation, QuantableOperation
from esp_ppq.IR.quantize import QuantableGraph
from esp_ppq.quantization.algorithm.training import *
from esp_ppq.quantization.measure import torch_mean_square_error, torch_snr_error
from esp_ppq.quantization.qfunction.linear import PPQLinearQuantFunction

from .base import QuantizationOptimizationPass


class TrainingBasedPass(QuantizationOptimizationPass):
    """Training Based Pass is a basic class that provides necessary function
    for all training optimizition passes. Optimization will be more stable and
    accurate with functions provided by this pass. (Might be a little slower).

    This pass will collect result of interested outputs after optimization and
        check if the optimized result has a lower SNR. If so, the optimization will be
        accepted, layer weight will be updated, otherwise optimization will be rejected and
        takes no effects.

    Choose interested_outputs carefully, cause we compare loss only with those output variables.
        If interested_outputs is None, all graph output variables will be chosen.

    YOUR SHOULD NOTICE THAT SNR REFERS TO: POWER OF NOISE / POWER OF SIGNAL IN esp_ppq.

    Args:
        QuantizationOptimizationPass ([type]): [description]
    """

    def __init__(
        self, name: str = 'Default Quanzation Optim', interested_outputs: List[str] = None, verbose: bool = True
    ) -> None:
        self._loss_fn = torch_snr_error
        self._interested_outputs = interested_outputs
        self._checkpoints = {}
        self._verbose = verbose
        self._quant_state_recorder = {}
        super().__init__(name=name)

    @empty_ppq_cache
    def initialize_checkpoints(
        self, graph: BaseGraph, executor: BaseGraphExecutor, dataloader: Iterable, collate_fn: Callable
    ):
        """
        Establish a series of network checkpoints with your network.
            Checkpoint is a data structure that helps us compare quant results and fp32 results.
        Args:
            graph (BaseGraph): [description]
            executor (BaseGraphExecutor): [description]
            dataloader (Iterable): [description]
            collate_fn (Callable): [description]

        Raises:
            PermissionError: [description]
        """
        for operation in graph.operations.values():
            if isinstance(operation, QuantableOperation):
                for cfg, var in operation.config_with_variable:
                    if cfg.state in {QuantizationStates.BAKED, QuantizationStates.PASSIVE_BAKED}:
                        raise PermissionError(
                            'Can not initialize checkpoints when weight value is baked. '
                            f'Variable {var.name} has a baked value.'
                        )

        if self._interested_outputs is None or len(self._interested_outputs) == 0:
            self._interested_outputs = [name for name in graph.outputs]

        for name in self._interested_outputs:
            self._checkpoints[name] = FinetuneCheckPoint(variable=name)

        # dequantize graph, collect references
        for op in graph.operations.values():
            if isinstance(op, QuantableOperation):
                op.dequantize()

        for data in tqdm(dataloader, desc='Collecting Referecens'):
            if collate_fn is not None:
                data = collate_fn(data)
            outputs = executor.forward(inputs=data, output_names=self._interested_outputs)
            for name, output in zip(self._interested_outputs, outputs):
                ckpt = self._checkpoints[name]
                assert isinstance(ckpt, FinetuneCheckPoint)
                ckpt.push(tensor=output, is_reference=True)

        # restore quantization state:
        for op in graph.operations.values():
            if isinstance(op, QuantableOperation):
                op.restore_quantize_state()

        # update state
        verbose, self._verbose = self._verbose, False
        self.check(executor=executor, dataloader=dataloader, collate_fn=collate_fn)
        self._verbose = verbose

    def check(self, executor: BaseGraphExecutor, dataloader: Iterable, collate_fn: Callable):
        """Check quantization error with a given dataloader with current
        checkpoints. Return whether quantization error is lower than before.

        Args:
            executor (BaseGraphExecutor): [description]
            dataloader (Iterable): [description]
            collate_fn (Callable): [description]

        Returns:
            [type]: [description]
        """

        # step - 1, collecting data
        for data in dataloader:
            if collate_fn is not None:
                data = collate_fn(data)
            outputs = executor.forward(inputs=data, output_names=self._interested_outputs)
            for name, output in zip(self._interested_outputs, outputs):
                self._checkpoints[name].push(tensor=output, is_reference=False)

        # step - 2, calculating loss
        losses = []
        for name in self._interested_outputs:
            ckpt = self._checkpoints[name]
            assert isinstance(ckpt, FinetuneCheckPoint)
            qt_out, fp_out = ckpt.pop()
            qt_out = torch.cat([tensor for tensor in qt_out])
            fp_out = torch.cat([tensor for tensor in fp_out])
            losses.append(self._loss_fn(y_pred=qt_out, y_real=fp_out).item())
            ckpt.clear()

        # step - 3, comparing loss
        loss_now, loss_old = sum(losses), sum([ckpt.best_loss for ckpt in self._checkpoints.values()])
        loss_now, loss_old = loss_now / len(losses), loss_old / len(losses)
        if self._verbose:
            print(f'NOISE-SIGNAL RATIO: {loss_old * 100:.4f}% -> {loss_now * 100:.4f}%.')

        # if there is a loss drop, update all losses.
        if loss_old > (loss_now * CHECKPOINT_TOLERANCE):
            for idx, name in enumerate(self._interested_outputs):
                ckpt = self._checkpoints[name]
                assert isinstance(ckpt, FinetuneCheckPoint)
                ckpt.best_loss = losses[idx]
            return True
        return False

    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        raise NotImplementedError(
            'Can not invoke this function. '
            'Please inherit this class and give an implementation to override this function.'
        )

    @empty_ppq_cache
    def enable_block_gradient(self, block: TrainableBlock):
        """
        Make all tensors inside a given block to be trainable(requres_grad = True)
        Both quantization scale and weight itself are going to be trained in training procedure

        Args:
            block (TrainableBlock): _description_
        """
        for op in block.rps:
            for var in op.inputs + op.outputs:
                if var.is_parameter and isinstance(var.value, torch.Tensor):
                    # PATCH 2022 08 01 Clip op can not be train
                    if op.type == 'Clip':
                        continue
                    if var.value.dtype == torch.float:
                        var.value.requires_grad = True
            if isinstance(op, QuantableOperation):
                for cfg, _ in op.config_with_variable:
                    if isinstance(cfg.scale, torch.Tensor):
                        cfg.scale.requires_grad = True

    @empty_ppq_cache
    def disable_block_gradient(self, block: TrainableBlock):
        for op in block.rps:
            for var in op.inputs + op.outputs:
                if var.is_parameter and isinstance(var.value, torch.Tensor):
                    if var.value.is_leaf:
                        var.value.requires_grad = False
                        var.value._grad = None
            if isinstance(op, QuantableOperation):
                for cfg, _ in op.config_with_variable:
                    if isinstance(cfg.scale, torch.Tensor):
                        if cfg.scale.is_leaf:
                            cfg.scale.requires_grad = False
                            cfg.scale._grad = None

    def split_graph_into_blocks(
        self,
        graph: BaseGraph,
        executing_order: List[Operation],
        blocksize: int = None,
        overlap: bool = False,
        interested_layers: List[str] = None,
    ) -> List[TrainableBlock]:
        """block construction function for training-based algorithms, if
        `block_limit` is not specified, block grandularity will be controlled by
        the default value OPTIM_ADVOPT_GRAPH_MAXSIZE specified in esp_ppq.core.common.

        Args:
            graph (BaseGraph): ppq ir graph
            executing_order (List[Operation]): topo search order
            block_limit (int, optional): controls maximum depth of a block. Defaults to None.

        Returns:
            List[TrainableBlock]: list of all partitioned blocks
        """
        if blocksize is None:
            blocksize = OPTIM_ADVOPT_GRAPH_MAXDEPTH
        visited_ops, blocks = set(), []
        block_builder = BlockBuilder(graph=graph, topo_order=executing_order)

        for op in graph.operations.values():
            # start from computing op
            if op in visited_ops and overlap is False:
                continue
            if isinstance(op, QuantableOperation) and op.is_computing_op:
                block = block_builder.build(op, blocksize)
                # by default blocks are exclusive from each other
                for op in block.rps:
                    visited_ops.add(op)
                blocks.append(block)

        ret = []
        if interested_layers is None or len(interested_layers) == 0:
            ret = blocks  # if no interested_layers, finetune all.
        else:
            for candidate in blocks:
                assert isinstance(candidate, TrainableBlock)
                if any([op.name in interested_layers for op in candidate.rps]):
                    ret.append(candidate)
        return ret

    def collect(
        self,
        graph: BaseGraph,
        block: TrainableBlock,
        executor: TorchExecutor,
        dataloader: Iterable,
        collate_fn: Callable,
        collecting_device: str,
        steps: int = None,
        expire_device: str = 'cpu',
    ) -> Tuple[List[Dict[str, torch.Tensor]], List[Dict[str, torch.Tensor]]]:
        """
        Collect training data for given block.
        This function will collect fp32 output and quantized input data by
            executing your graph twice.
        For collecting fp32 output, all related operations will be dequantized.
        For collecting quantized input, all related operations' quantization state will be restored.

        collecting device declares where cache to be stored:
            executor - store cache to executor device.(default)
            cpu      - store cache to system memory.
            cuda     - store cache to gpu memory.(2x speed up)
            disk     - not implemented.

        Args:
            block (TrainableBlock): _description_
            executor (TorchExecutor): _description_
            dataloader (Iterable): _description_
            collate_fn (Callable): _description_
            collecting_device (str): _description_

        Returns:
            _type_: _description_
        """

        def cache_fn(data: torch.Tensor):
            # TODO move this function to esp_ppq.core.IO
            if not isinstance(data, torch.Tensor):
                raise TypeError(
                    'Unexpected Type of value, Except network output to be torch.Tensor, '
                    f'however {type(data)} was given.'
                )
            if collecting_device == 'cpu':
                data = data.cpu()
            if collecting_device == 'cuda':
                data = data.cuda()
            # TODO restrict collecting device.
            return data

        with torch.no_grad():
            try:
                if len(dataloader) > 1024:
                    ppq_warning(
                        'Large finetuning dataset detected(>1024). '
                        'You are suppose to prepare a smaller dataset for finetuning. '
                        'Large dataset might cause system out of memory, '
                        'cause all data are cache in memory.'
                    )
            except Exception as e:
                pass  # dataloader has no __len__

            quant_graph = QuantableGraph(graph)  # helper class
            fp_outputs, qt_inputs = [], []

            cur_iter = 0
            # dequantize graph, collect fp32 outputs
            quant_graph.dequantize_graph(expire_device=expire_device)
            for data in dataloader:
                if collate_fn is not None:
                    data = collate_fn(data)
                fp_output = executor.forward(data, [var.name for var in block.ep.outputs])
                fp_output = {var.name: cache_fn(data) for data, var in zip(fp_output, block.ep.outputs)}
                fp_outputs.append(fp_output)
                cur_iter += 1
                if steps is not None and cur_iter > steps:
                    break

            cur_iter = 0
            # restore quantization state, collect quant inputs
            quant_graph.restore_quantize_state(expire_device=expire_device)
            for data in dataloader:
                if collate_fn is not None:
                    data = collate_fn(data)
                # PATCH 20220829, 有些 computing op 权重并非定值
                non_constant_input = [var for var in block.sp.inputs if not var.is_parameter]
                qt_input = executor.forward(data, [var.name for var in non_constant_input])
                qt_input = {var.name: cache_fn(value) for var, value in zip(non_constant_input, qt_input)}
                qt_inputs.append(qt_input)
                cur_iter += 1
                if steps is not None and cur_iter > steps:
                    break

        return qt_inputs, fp_outputs

    def compute_block_loss(
        self,
        block: TrainableBlock,
        qt_inputs: List[Dict[str, torch.Tensor]],
        fp_outputs: List[Dict[str, torch.Tensor]],
        executor: TorchExecutor,
        loss_fn: Callable = torch_mean_square_error,
    ) -> float:
        """
        loss computing for fp32 and quantized graph outputs, used
        in multiple training-based algorithms below

        Args:
            output_names (List[str]): output variable names
            graph (BaseGraph): ppq ir graph
            dataloader (Iterable): calibration dataloader
            collate_fn (Callable): batch collate func
            executor (TorchExecutor): ppq torch executor
            loss_fn (Callable, optional): loss computing func. Defaults to torch_mean_square_error.
        Returns:
            Dict[str, float]: loss dict for variables specified in `output_names`
        """
        with torch.no_grad():
            losses = {var.name: 0.0 for var in block.ep.outputs}
            output_names = [var.name for var in block.ep.outputs]

            for qt_input, fp_output in zip(qt_inputs, fp_outputs):
                feed_dict = {k: v.to(executor._device) for k, v in qt_input.items()}

                qt_output = executor.partial_graph_forward(
                    operations=block.rps, feed_dict=feed_dict, output_names=output_names
                )

                for name, quant_output in zip(output_names, qt_output):
                    batch_loss = loss_fn(quant_output, fp_output[name].to(executor._device))
                    losses[name] += batch_loss.detach().item()

            for name in losses:
                losses[name] /= len(qt_inputs)
        return sum([v for v in losses.values()])


class BiasCorrectionPass(TrainingBasedPass):
    """
    ## Bias Correction Optimization Pass(Bias 校准过程)

    Bias correction is the process of shifting quantized model outputs to account for their statistical errors.

    Network quantization will bring some error(noise) to the result. To improve the accuracy of a quantized model,
    we can correct the network by adding an extra term on bias in order to make the output has zero expectation.

    Bias correction is used to eliminate bias error, generally it will take a few minutes to correct all bias terms.

    For those layers have no bias, Bias Correction Optimization will skip them directly.

        let: Y = WX + b

        Quant(Y) = Qunat(W) Quant(X) + b

        bias_error = reduce_mean(Y - Quant(Y))

        This pass will correct bias with: b = b + bias_error

    ### Parameters:

    * interested_layers(List[str]):

            A list of operation names, only the layers listed in this parameter will be processed.

            If interested_layers is None, all layers will be processed.

    * steps(int)

            Forward steps for collecting bias error,
            a large value of this parameter means more data will be collected so
            the bias error will be estimated better, while it takes more time.

            Usually 8 ~ 32 step is enough in most cases.

    * block_size(int)

            Bias Correction Optimization will split your graph into blocks,
            bias error will be collected and corrected block by block.

            A large block size will greatly reduce running time of this optimization,
            while it might give an unstable result when blocksize is too large.

            By default this value is set to 4, to have the best result of optimization, you are recommended to set blocksize = 1.

    * loss_fn(Callable)

            A function that used to measure the loss after optimization.

            Bias Correction Optimization is a training-based pass,
            we will check the loss at the end of block optimization.

            If the optimization created worsen result, the optimization result will be drop.

    ### Usage:

    Bias Correction Optimization Pass should be invoked after Runtime Calibration Pass.

    This pass is included in PPQ Quantization Setting, you can calling this optimization by:

        setting = QuantizationSettingFactory.default_setting()

        setting.bias_correct = True

        # calling esp_ppq.api.quantize_onnx_model function with this setting.
        ir = quantize_torch_model(
        model=model, calib_dataloader=load_calibration_dataset(), setting=setting,
        platform=TargetPlatform.PPL_CUDA_INT8, calib_steps=8, input_shape=INPUT_SHAPE,
        collate_fn=collate_fn)

    You can manually create this optimization by:

        from esp_ppq import BiasCorrectionPass

        optim = BiasCorrectionPass()

    ### Version:

    Require PPQ 0.5.2 +

    Interface changed since PPQ 0.6.5
    """

    def __init__(
        self, interested_layers: List[str] = [], collecting_device: str = 'cuda', steps: int = 32, block_size: int = 1
    ) -> None:
        super().__init__(name='PPQ Bias Correction Pass')
        self.interested_layers = interested_layers
        self.steps = steps
        self.block_size = block_size
        self.collecting_device = collecting_device
        self.loss_fn = torch_mean_square_error

    @empty_ppq_cache
    def correct_bias(
        self,
        qt_inputs: List[Dict[str, torch.Tensor]],
        fp_outputs: List[Dict[str, torch.Tensor]],
        block: TrainableBlock,
        executor: TorchExecutor,
        graph: BaseGraph,
    ) -> Tuple[float, float]:
        def collect_bias(output: torch.Tensor, op_type: str) -> torch.Tensor:
            if output.ndim < 1:
                raise ValueError('Forward value has an unexpected dimension.')
            if op_type in {'Conv', 'ConvTranspose'}:
                # for convolution layer, bias always been added on axis 1
                reduce_dims = [i for i in range(output.ndim) if i != 1]
                return torch.mean(output, dim=reduce_dims).unsqueeze(0)
            elif op_type in {'Gemm'}:
                # for convolution layer, bias always been added on axis -1
                reduce_dims = [i for i in range(output.ndim) if i != (output.ndim - 1)]
                return torch.mean(output, dim=(0,)).unsqueeze(0)
            else:
                raise TypeError(f'Unsupported Operation type: {op_type}')

        def dequantize_block(block: TrainableBlock):
            for op in block.rps:
                if isinstance(op, QuantableOperation):
                    op.dequantize()

        def restore_block_quantize_state(block: TrainableBlock):
            for op in block.rps:
                if isinstance(op, QuantableOperation):
                    op.restore_quantize_state()

        with torch.no_grad():
            # record pre training loss.
            pre_loss = self.compute_block_loss(
                block=block, qt_inputs=qt_inputs, fp_outputs=fp_outputs, executor=executor, loss_fn=self.loss_fn
            )

            bias_cloned, interested_outputs, fp_cache, qt_cache = {}, [], defaultdict(list), defaultdict(list)
            for op in block.rps:
                op_check = op.type in {'Conv', 'ConvTranspose', 'Gemm'} and len(op.inputs) == 3
                bias_check = op.inputs[-1].is_parameter
                type_check = isinstance(op.inputs[-1].value, torch.Tensor)

                if op_check and bias_check and type_check:
                    bias_cloned[op.name] = op.inputs[-1].value.clone()
                    interested_outputs.append(op.outputs[0].name)

            # Phrase 1, collect fp32 output
            dequantize_block(block)
            for qt_input, _ in tqdm(zip(qt_inputs, fp_outputs), '# Bias Correction Phrase 1'):
                feed_dict = {k: v.to(executor._device) for k, v in qt_input.items()}

                outputs = executor.partial_graph_forward(
                    operations=block.rps, feed_dict=feed_dict, output_names=interested_outputs
                )

                for name, value in zip(interested_outputs, outputs):
                    source_op = graph.variables[name].source_op
                    fp_cache[name].append(collect_bias(value, source_op.type))

            # Phrase 2, collect quant output
            restore_block_quantize_state(block)
            for qt_input, _ in tqdm(zip(qt_inputs, fp_outputs), '# Bias Correction Phrase 2'):
                feed_dict = {k: v.to(executor._device) for k, v in qt_input.items()}

                outputs = executor.partial_graph_forward(
                    operations=block.rps, feed_dict=feed_dict, output_names=interested_outputs
                )

                for name, value in zip(interested_outputs, outputs):
                    source_op = graph.variables[name].source_op
                    qt_cache[name].append(collect_bias(value, source_op.type))

            # correct bias error
            for name in fp_cache:
                DC_term_fp = fp_cache[name]
                DC_term_qt = qt_cache[name]

                if len(DC_term_fp) == 0 or len(DC_term_qt) == 0:
                    raise ValueError('Bias correction failed, No data was collected.')
                DC_term_fp = torch.mean(torch.cat(DC_term_fp, axis=0), dim=0)
                DC_term_qt = torch.mean(torch.cat(DC_term_qt, axis=0), dim=0)
                bias_error = DC_term_fp - DC_term_qt

                source_op = graph.variables[name].source_op
                source_op.inputs[-1].value += bias_error

            # record pre training loss.
            post_loss = self.compute_block_loss(
                block=block, qt_inputs=qt_inputs, fp_outputs=fp_outputs, executor=executor, loss_fn=self.loss_fn
            )

            # loss check
            if post_loss > pre_loss:
                for name, value in bias_cloned.items():
                    op = graph.operations[name]
                    op.inputs[-1].value.copy_(value)
                post_loss = pre_loss

        return pre_loss, post_loss

    @empty_ppq_cache
    def optimize(
        self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, collate_fn: Callable, **kwargs
    ) -> None:
        blocks = self.split_graph_into_blocks(
            graph=graph,
            executing_order=executor._executing_order,
            blocksize=self.block_size,
            interested_layers=self.interested_layers,
        )

        # ready for finetuning, print information.
        print('')
        print('Check following parameters:')
        print(f'Interested Layers:         {self.interested_layers}')
        print(f'Num of blocks:             {len(blocks)}')
        print(f'Steps:                     {self.steps}')
        print(f'collecting_device:         {self.collecting_device}')
        print('')  # blank line

        # do per-block finetune
        for block_idx, block in enumerate(blocks):
            qt_inputs, fp_outputs = self.collect(
                graph=graph,
                block=block,
                executor=executor,
                dataloader=dataloader,
                collate_fn=collate_fn,
                collecting_device=self.collecting_device,
            )

            print(f'# Block [{block_idx + 1} / {len(blocks)}]: [{block.sp.name} -> {block.ep.name}]')
            pre_loss, post_loss = self.correct_bias(
                qt_inputs=qt_inputs, fp_outputs=fp_outputs, block=block, executor=executor, graph=graph
            )
            print(f'# Tuning Finished  : ({pre_loss:.4f} -> {min(pre_loss, post_loss):.4f}) [Block Loss]')
            print('')  # blank line


class LearnedStepSizePass(TrainingBasedPass):
    """
    ## Learned Step Size Pass(网络微调过程-LSQ)

    Learned Step Size optimization, a training-based optimization pass that tunes weights and scales for high precision quantization.

    [This method is proposed by Steven K. Esser] (https://arxiv.org/pdf/1902.08153.pdf)

    This is an alternative version of LSQ, this pass will split your graph into multiple trainable blocks, each blocks will be trained separately.
    Warning: PPQ Learned Step Size minimize only the output loss of each block, which means after training the internal results probably goes far away from original.

    PPQ Learned Step Size optimization requires 256 ~ 2048 samples for finetuning your network, while the data label is not necessary. All training data are cache in GPU memory or CPU memory for acceleration.

    The training loss will be computed as:

        let: Y = WX + b

        Quant(Y, scale_Y) = Quant(W, scale_W) Quant(X, scale_X) + b

        loss = loss_func(Y, Quant(Y, scale_Y)) # loss between fp output and int8 output, that is why we do not need labeled data.

    The formula of calculating the derivatives of y and scale_Y:

        if y > scale_Y * -128 and y < scale_Y * 127:

        dQuant(y, scale_Y)/dy       = dQuant(y, scale_Y)

        dQuant(y, scale_Y)/dscale_Y = Quant(y, scale_Y) - y

        if y < scale_Y * -128:

        dQuant(y, scale_Y)/dy       = 0

        dQuant(y, scale_Y)/dscale_Y = -128

        if y > scale_Y * 127:

        dQuant(y, scale_Y)/dy       = 0

        dQuant(y, scale_Y)/dscale_Y = 127

    ### Parameters:

    * interested_layers(List[str]):

            A list of operation names, only the layers listed in this parameter will be trained.

            If interested_layers is None, all layers(conv and gemm) will be trained.

    * steps(int)

            Training steps for finetuning your network, default is 500.

    * block_size(int)

            PPQ Learned Step Size optimization split your graph into blocks at first,
            each block will be finetuned separately.

            A large block size will greatly reduce running time of this optimization,
            while it might give an unstable result when blocksize is too large.

            By default this value is set to 4.

    * is_scale_trainable(bool)

            If is_scale_trainable = False, optimization will not apply to scales, only network parameters will be tuned.

            Scale is trainable when all the following conditions are fulfilled:

                1. scale is valid
                2. corresponding tensor quantization config state is active
                3. do not have POWER_OF_2 policy
                4. is_scale_trainable = True

    * gamma(float)

            A regularization term for minimize the distance of Y and Quant(Y)

            If gamma is not 0, loss = loss_func(Y, Quant(Y, scale_Y)) + MSE(Y, Quant(Y)) * gamma

            Default is 0

    * lr(float)

            Learning rate, Default is 5e-5

    * collecting_device(str)

            String that representing the device on which cache data is or will be allocated.

            Can be cpu, cuda, disk

    * loss_fn(Callable)

            A function that used to measure the loss after optimization.

            LSQ is a training-based pass,
            we will check the loss at the end of block optimization.

            If the result goes worsen, optimized weights and scales will be drop.

    ### Usage:

    LSQ Optimization Pass should be invoked after Runtime Calibration Pass.

    This pass is inclueded in PPQ Quantization Setting, you can calling this optimization by:

        setting = QuantizationSettingFactory.default_setting()

        setting.lsq_optimization = True

        # calling esp_ppq.api.quantize_onnx_model function with this setting.
        ir = quantize_torch_model(
        model=model, calib_dataloader=load_calibration_dataset(), setting=setting,
        platform=TargetPlatform.PPL_CUDA_INT8, calib_steps=8, input_shape=INPUT_SHAPE,
        collate_fn=collate_fn)

    You can manually create this optimization by:

        from esp_ppq import LearnedStepSizePass

        optim = LearnedStepSizePass()


    ### Block-partitioning Algorithm

    PPQ use Block-partitioning algorithm for spliting your graph into blocks, this method is based on graph theory.

    Parameter block_size will controls the maximum size of created blocks.

    If block_size = 1, then each block will contains exactly 1 layer within it, blockwise optimization will degenerate to layerwise optimization.

    If block_size is set to a large value, training progress will be unstable since batchnorm layers have been merged at first.

    ### Version:

    Require PPQ 0.6.2 +

    Interface changed since PPQ 0.6.5
    """

    def __init__(
        self,
        name: str = 'PPQ LSQ Optimization',
        interested_layers: List[str] = [],
        steps: int = 500,
        gamma: float = 0.0,
        is_scale_trainable: bool = True,
        lr: float = 5e-5,
        block_size: int = 5,
        expire_device: str = 'cpu',
        collecting_device: str = 'cuda',
        loss_fn: Callable = torch_mean_square_error,
        optimizer: Any = None,
    ) -> None:
        super().__init__(name=name)
        self.interested_layers = interested_layers
        self.collecting_device = collecting_device
        self.expire_device = expire_device
        self.is_scale_trainable = is_scale_trainable
        self.block_size = block_size
        self.loss_fn = loss_fn
        self.gamma = gamma
        self.steps = steps
        self.lr = lr
        self.optimizer = optimizer

    def finetune(
        self,
        steps: int,
        learning_rate: float,
        block: TrainableBlock,
        executor: TorchExecutor,
        qt_inputs: List[Dict[str, torch.Tensor]],
        fp_outputs: List[Dict[str, torch.Tensor]],
        optimizer: torch.optim.Optimizer = None,
        scheduler: object = None,
    ) -> Tuple[float, float]:
        # step - 1: enable gradient for training.
        self.enable_block_gradient(block)

        # record pre training loss.
        pre_loss = self.compute_block_loss(
            block=block, qt_inputs=qt_inputs, fp_outputs=fp_outputs, executor=executor, loss_fn=self.loss_fn
        )

        # collect trainable params
        trainable_params, delegators = [], {}
        trainable_scales = []
        for op in block.rps:
            if not isinstance(op, QuantableOperation):
                continue

            if op.is_computing_op or op.type in {'Add', 'Mul'}:  # 独立的 Add Mul 算子也可以训练
                for var in op.inputs:
                    if var.is_parameter:
                        trainable_params.append(var.value)

            # register quant delegator
            for cfg, var in op.config_with_variable:
                if cfg.state in {QuantizationStates.ACTIVATED, QuantizationStates.PASSIVE}:
                    delegator = LSQDelegator(config=cfg, var=var)
                    trainable_scales.extend(delegator.trainable_tensors())
                    executor.register_quantize_delegate(config=cfg, delegator=delegator)
                    delegators[cfg] = delegator

        # check if empty.
        tensors = [tensor for tensor in trainable_params + trainable_scales if tensor.requires_grad]
        tensors = set(tensors)  # remove duplicated tensor
        if len(tensors) == 0:
            for cfg, delegator in delegators.items():
                executor.remove_quantize_delegate(config=cfg)
            return 0, 0

        # initilize optimizer.
        if self.optimizer is None:
            optimizer = torch.optim.Adam(tensors, lr=learning_rate)
        else:
            optimizer = self.optimizer(tensors, lr=learning_rate)

        dataset_length = len(qt_inputs)
        if dataset_length == 0:
            raise ValueError('Dataset is empty.')

        # step 2 - training procedure
        for idx in tqdm(range(steps), desc='# Tuning Procedure '):
            qt_input, fp_output = qt_inputs[idx % dataset_length], fp_outputs[idx % dataset_length]

            # forward
            optimizer.zero_grad()
            feed_dict = {k: v.to(executor._device) for k, v in qt_input.items()}
            output_names = [name for name in fp_output]

            qt_output = executor.partial_graph_forward(
                operations=block.rps, feed_dict=feed_dict, output_names=output_names
            )

            # compute loss
            loss = 0.0
            for idx, name in enumerate(output_names):
                loss += self.loss_fn(qt_output[idx], fp_output[name].to(executor._device))

            for op in block.rps:
                if self.gamma == 0:
                    continue
                if isinstance(op, QuantableOperation) and op.is_computing_op:
                    weight = op.inputs[1].value
                    wconfig = op.config.input_quantization_config[1]
                    loss += torch_mean_square_error(weight, PPQLinearQuantFunction(weight, wconfig)) * self.gamma

            # backward from loss
            assert isinstance(loss, torch.Tensor)
            loss.backward()
            optimizer.step()
            if scheduler is not None:
                scheduler.step()

        # step - 3: record post training loss
        post_loss = self.compute_block_loss(
            block=block, qt_inputs=qt_inputs, fp_outputs=fp_outputs, executor=executor, loss_fn=self.loss_fn
        )
        # check and withdraw
        if post_loss > pre_loss:
            for cfg, delegator in delegators.items():
                delegator.withdraw()

        for cfg, delegator in delegators.items():
            delegator.finalize()
            executor.remove_quantize_delegate(config=cfg)

        # disable gradient for evaluation.
        self.disable_block_gradient(block)

        # clear cache
        torch.cuda.empty_cache()
        return pre_loss, post_loss

    def optimize(
        self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, collate_fn: Callable, **kwargs
    ) -> None:
        blocks = self.split_graph_into_blocks(
            graph=graph,
            executing_order=executor._executing_order,
            blocksize=self.block_size,
            interested_layers=self.interested_layers,
        )

        # ready for finetuning, print information.
        print('')
        print('Check following parameters:')
        print(f'Is Scale Trainable:        {self.is_scale_trainable}')
        print(f'Interested Layers:         {self.interested_layers}')
        print(f'Collecting Device:         {self.collecting_device}')
        print(f'Num of blocks:             {len(blocks)}')
        print(f'Learning Rate:             {self.lr}')
        print(f'Steps:                     {self.steps}')
        print(f'Gamma:                     {self.gamma}')
        print('')  # blank line

        # do per-block finetune
        for block_idx, block in enumerate(blocks):
            # collect data for training
            qt_inputs, fp_outputs = self.collect(
                graph=graph,
                block=block,
                executor=executor,
                dataloader=dataloader,
                collate_fn=collate_fn,
                collecting_device=self.collecting_device,
            )

            print(f'# Block [{block_idx + 1} / {len(blocks)}]: [{block.sp.name} -> {block.ep.name}]')
            pre_loss, post_loss = self.finetune(
                steps=self.steps,
                learning_rate=self.lr,
                block=block,
                qt_inputs=qt_inputs,
                fp_outputs=fp_outputs,
                executor=executor,
            )
            print(f'# Tuning Finished  : ({pre_loss:.4f} -> {min(pre_loss, post_loss):.4f}) [Block Loss]')
            print('')  # blank line


class RoundTuningPass(TrainingBasedPass):
    def __init__(
        self,
        interested_layers: List[str] = [],
        steps: int = 500,
        lr: float = 1e-4,
        block_size: int = 5,
        expire_device: str = 'cpu',
        collecting_device: str = 'cuda',
        optimizer: Any = None,
    ) -> None:
        """This method trains the weights of neural networks to make them more suitable for quantization compression.
        Similar to Adaround, this method restricts the range of weight training within one scale [-scale, +scale],
        so this method can also be called round mode adjustment.

        Compared to Adaround, this method uses a more direct and simple way to train weights,
        and its training effect is similar to Adaround.

        Reference: Up or Down? Adaptive Rounding for Post-Training Quantization

        Parameter:
            interested_layers: specifies which layers of the neural network need to be trained.
                It is a list containing the names of all the layers that need to be trained.

            collecting_device: specifies where the data cache is stored during training,
                supporting caching to main memory (CPU) or caching to graphics memory (CUDA).
                Note that the dataset used for training will be fully cached.

            block_size: used to determine the maximum depth of the subgraph
                during the training process of the neural network.

            steps: training steps.

            lr: learning rate.

            optimizer: training optimizer(default: Adam).
        """
        super().__init__(name='PPQ Rounding Tuning Pass')
        self.interested_layers = interested_layers
        self.collecting_device = collecting_device
        self.expire_device = expire_device
        self.block_size = block_size
        self.steps = steps
        self.lr = lr
        self.optimizer = optimizer
        self.loss_fn = torch.nn.MSELoss()

    def finetune(
        self,
        steps: int,
        learning_rate: float,
        block: TrainableBlock,
        executor: TorchExecutor,
        qt_inputs: List[Dict[str, torch.Tensor]],
        fp_outputs: List[Dict[str, torch.Tensor]],
        optimizer: torch.optim.Optimizer = None,
        scheduler: object = None,
    ) -> float:
        # step - 1: enable gradient for training.
        self.enable_block_gradient(block)

        # record pre training loss.
        pre_loss = self.compute_block_loss(block=block, qt_inputs=qt_inputs, fp_outputs=fp_outputs, executor=executor)

        # collect trainable params
        trainable_params, delegators = [], {}
        for op in block.rps:
            if not isinstance(op, QuantableOperation):
                continue
            if op.num_of_parameter == 0:
                continue

            if op.type in {'Gemm', 'MatMul', 'ConvTranspose', 'PPQBiasFusedMatMul', 'Conv'}:
                if op.inputs[1].is_parameter:
                    cfg, var = op.config.input_quantization_config[1], op.inputs[1]
                    delegator = RoundTruningDelegator(config=cfg, var=var)
                    trainable_params.append(delegator._rounding)
                    executor.register_quantize_delegate(config=cfg, delegator=delegator)
                    delegators[cfg] = delegator

                if op.num_of_input == 3 and op.inputs[-1].is_parameter:
                    op.inputs[-1].value.requires_grad = True  # clear it.
                    trainable_params.append(op.inputs[-1].value)

        # check if empty.
        tensors = [tensor for tensor in trainable_params if tensor.requires_grad]
        tensors = set(tensors)  # remove duplicated tensor
        if len(tensors) == 0:
            for cfg, delegator in delegators.items():
                executor.remove_quantize_delegate(config=cfg)
            return 0, 0

        # initilize optimizer.
        if self.optimizer is None:
            optimizer = torch.optim.Adam(tensors, lr=learning_rate)
        else:
            optimizer = self.optimizer(tensors, lr=learning_rate)

        dataset_length = len(qt_inputs)
        if dataset_length == 0:
            raise ValueError('Dataset is empty.')

        # step 2 - training procedure
        for idx in tqdm(range(steps), desc='# Tuning Procedure '):
            qt_input, fp_output = qt_inputs[idx % dataset_length], fp_outputs[idx % dataset_length]

            # forward
            optimizer.zero_grad()
            feed_dict = {k: v.to(executor._device) for k, v in qt_input.items()}
            output_names = [name for name in fp_output]

            qt_output = executor.partial_graph_forward(
                operations=block.rps, feed_dict=feed_dict, output_names=output_names
            )

            # compute loss
            loss = 0.0
            for idx, name in enumerate(output_names):
                loss += self.loss_fn(qt_output[idx], fp_output[name].to(executor._device))

            # backward from loss
            assert isinstance(loss, torch.Tensor)
            loss.backward()
            optimizer.step()
            if scheduler is not None:
                scheduler.step()

        # step - 3: record post training loss
        post_loss = self.compute_block_loss(
            block=block, qt_inputs=qt_inputs, fp_outputs=fp_outputs, executor=executor, loss_fn=self.loss_fn
        )
        # check and withdraw
        if post_loss > pre_loss:
            for cfg, delegator in delegators.items():
                delegator.withdraw()

        for cfg, delegator in delegators.items():
            delegator.finalize()
            executor.remove_quantize_delegate(config=cfg)

        # disable gradient for evaluation.
        self.disable_block_gradient(block)

        # clear cache
        torch.cuda.empty_cache()
        return pre_loss, post_loss

    def optimize(
        self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, collate_fn: Callable, **kwargs
    ) -> None:
        blocks = self.split_graph_into_blocks(
            graph=graph,
            executing_order=executor._executing_order,
            blocksize=self.block_size,
            interested_layers=self.interested_layers,
        )

        # ready for finetuning, print information.
        print('')
        print('Check following parameters:')
        print(f'Interested Layers:         {self.interested_layers}')
        print(f'Collecting Device:         {self.collecting_device}')
        print(f'Num of blocks:             {len(blocks)}')
        print(f'Learning Rate:             {self.lr}')
        print(f'Steps:                     {self.steps}')
        print('')  # blank line

        # do per-block finetune
        for block_idx, block in enumerate(blocks):
            # collect data for training
            qt_inputs, fp_outputs = self.collect(
                graph=graph,
                block=block,
                executor=executor,
                dataloader=dataloader,
                collate_fn=collate_fn,
                collecting_device=self.collecting_device,
            )

            print(f'# Block [{block_idx + 1} / {len(blocks)}]: [{block.sp.name} -> {block.ep.name}]')
            pre_loss, post_loss = self.finetune(
                steps=self.steps,
                learning_rate=self.lr,
                block=block,
                qt_inputs=qt_inputs,
                fp_outputs=fp_outputs,
                executor=executor,
            )
            print(f'# Tuning Finished  : ({pre_loss:.4f} -> {min(pre_loss, post_loss):.4f}) [Block Loss]')
            print('')  # blank line



import onnx
from onnx import helper, version_converter

# Preprocessing: load the model to be converted.
model_path = 'models/mbfill.onnx'
original_model = onnx.load(model_path)

# A full list of supported adapters can be found here:
# https://github.com/onnx/onnx/blob/main/onnx/version_converter.py#L21
# Apply the version conversion on the original model
converted_model = version_converter.convert_version(original_model, 11)

onnx.save(converted_model, 'models/mbfill_11.onnx')



import torch

import esp_ppq.lib as PFL
from esp_ppq import TargetPlatform, TorchExecutor, graphwise_error_analyse
from esp_ppq.api import ENABLE_CUDA_KERNEL
from esp_ppq.api.interface import load_onnx_graph
from esp_ppq.core.quant import QuantizationPolicy, QuantizationProperty, RoundingPolicy
from esp_ppq.quantization.optim import (
    LearnedStepSizePass,
    ParameterBakingPass,
    ParameterQuantizePass,
    QuantAlignmentPass,
    QuantizeFusionPass,
    QuantizeSimplifyPass,
    RuntimeCalibrationPass,
)

# ------------------------------------------------------------
# 在这个例子中我们将向你展示如何使用 INT8 量化一个 Yolo v6 模型
# 我们使用随机数据进行量化，这并不能得到好的量化结果。
# 在量化你的网络时，你应当使用真实数据和正确的预处理。
# ------------------------------------------------------------
graph = load_onnx_graph(onnx_import_file='yolov6s.onnx')
dataset = [torch.rand(size=[1, 3, 640, 640]) for _ in range(64)]

# -----------------------------------------------------------
# 我们将借助 PFL - PPQ Foundation Library, 即 PPQ 基础类库完成量化
# 这是 PPQ 自 0.6.6 以来推出的新的量化 api 接口，这一接口是提供给
# 算法工程师、部署工程师、以及芯片研发人员使用的，它更为灵活。
# 我们将手动使用 Quantizer 完成算子量化信息初始化, 并且手动完成模型的调度工作
# ------------------------------------------------------------
collate_fn = lambda x: x.cuda()

# ------------------------------------------------------------
# 在开始之前，我需要向你介绍量化器、量化信息以及调度表
# 量化信息在 PPQ 中是由 TensorQuantizationConfig(TQC) 进行描述的
# 这个结构体描述了我要如何去量化一个数据，其中包含了量化位宽、量化策略、
# 量化 Scale, offset 等内容。
# ------------------------------------------------------------
from esp_ppq import TensorQuantizationConfig as TQC

MyTQC = TQC(
    policy=QuantizationPolicy(
        QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
    ),
    rounding=RoundingPolicy.ROUND_HALF_EVEN,
    num_of_bits=8,
    quant_min=-128,
    quant_max=127,
    exponent_bits=0,
    channel_axis=None,
    observer_algorithm='minmax',
)
# ------------------------------------------------------------
# 作为示例，我们创建了一个 "线性" "对称" "Tensorwise" 的量化信息
# 这三者皆是该量化信息的 QuantizationPolicy 的一部分
# 同时要求该量化信息使用 ROUND_HALF_EVEN 方式进行取整
# 量化位宽为 8 bit，其中指数部分为 0 bit
# 量化上限为 127.0，下限则为 -128.0
# 这是一个 Tensorwise 的量化信息，因此 channel_axis = None
# observer_algorithm 表示在未来使用 minmax calibration 方法确定该量化信息的 scale

# 上述例子完成了该 TQC 的初始化，但并未真正启用该量化信息
# MyTQC.scale, MyTQC.offset 仍然为空，它们必须经过 calibration 才会具有有意义的值
# 并且他目前的状态 MyTQC.state 仍然是 Quantization.INITIAL，这意味着在计算时该 TQC 并不会参与运算。
# ------------------------------------------------------------

# ------------------------------------------------------------
# 接下来我们向你介绍量化器，这是 PPQ 中的一个核心类型
# 它的职责是为网络中所有处于量化区的算子初始化量化信息(TQC)
# PPQ 中实现了一堆不同的量化器，它们分别适配不同的情形
# 在这个例子中，我们分别创建了 TRT_INT8, GRAPHCORE_FP8, TRT_FP8 三种不同的量化器
# 由它们所生成的量化信息是不同的，为此你可以访问它们的源代码
# 位于 esp_ppq.quantization.quantizer 中，查看它们初始化量化信息的逻辑。
# ------------------------------------------------------------
_ = PFL.Quantizer(platform=TargetPlatform.TRT_FP8, graph=graph)  # 取得 TRT_FP8 所对应的量化器
_ = PFL.Quantizer(platform=TargetPlatform.GRAPHCORE_FP8, graph=graph)  # 取得 GRAPHCORE_FP8 所对应的量化器
quantizer = PFL.Quantizer(platform=TargetPlatform.TRT_INT8, graph=graph)  # 取得 TRT_INT8 所对应的量化器

# ------------------------------------------------------------
# 调度器是 PPQ 中另一核心类型，它负责切分计算图
# 在量化开始之前，你的计算图将被切分成可量化区域，以及不可量化区域
# 不可量化区域往往就是那些执行 Shape 推断的算子所构成的子图
# *** 量化器只为量化区的算子初始化量化信息 ***
# 调度信息将被写在算子的属性中，你可以通过 op.platform 来访问每一个算子的调度信息
# ------------------------------------------------------------
dispatching = PFL.Dispatcher(graph=graph).dispatch(  # 生成调度表
    quant_types=quantizer.quant_operation_types
)

for op in graph.operations.values():
    # quantize_operation - 为算子初始化量化信息，platform 传递了算子的调度信息
    # 如果你的算子被调度到 TargetPlatform.FP32 上，则该算子不量化
    # 你可以手动修改调度信息
    dispatching['Op1'] = TargetPlatform.FP32  # 将 Op1 强行送往非量化区
    dispatching['Op2'] = TargetPlatform.UNSPECIFIED  # 将 Op2 强行送往量化区

    # 你可能已经注意到了，我们并没有将 Op2 送往 TRT_INT8，而是将其送往 UNSPECIFIED 平台
    # 其含义是告诉量化器我们 "建议" 量化器对算子初始化量化信息，如果此时 Op2 不是量化器所支持的类型，则该算子仍然不会被量化
    # 但如果我们直接将 Op2 送往 TRT_INT8，不论如何该算子都将被量化
    quantizer.quantize_operation(op_name=op.name, platform=dispatching[op.name])

# ------------------------------------------------------------
# 在创建量化管线之前，我们需要初始化执行器，它用于模拟硬件并执行你的网络
# 请注意，执行器需要对网络结果进行分析并缓存分析结果，如果你的网络结构发生变化
# 你必须重新建立新的执行器。在上一步操作中，我们对算子进行了量化，这使得
# 普通的算子被量化算子替代，这一步操作将会改变网络结构。因此我们必须在其后建立执行器。
# ------------------------------------------------------------
executor = TorchExecutor(graph=graph, device='cuda')
executor.tracing_operation_meta(inputs=collate_fn(dataset[0]))
executor.load_graph(graph=graph)

# ------------------------------------------------------------
# 下面的过程将创建量化管线，它还是一个 PPQ 的核心类型
# 在 PPQ 中，模型的量化是由一个一个的量化过程(QuantizationOptimizationPass)完成的
# 量化管线 是 量化过程 的集合，在其中的量化过程将被逐个调用
# 从而实现对 TQC 中内容的修改，最终实现模型的量化
# 在这里我们为管线中添加了 7 个量化过程，分别处理不同的内容

# QuantizeSimplifyPass - 用于移除网络中的冗余量化信息
# QuantizeFusionPass - 用于调整量化信息状态，从而模拟推理图融合
# ParameterQuantizePass - 用于为模型中的所有参数执行 Calibration, 生成它们的 scale，并将对应 TQC 的状态调整为 ACTIVED
# RuntimeCalibrationPass - 用于为模型中的所有激活执行 Calibration, 生成它们的 scale，并将对应 TQC 的状态调整为 ACTIVED
# QuantAlignmentPass - 用于执行 concat, add, sum, sub, pooling 算子的定点对齐
# LearnedStepSizePass - 用于训练微调模型的权重，从而降低量化误差
# ParameterBakingPass - 用于执行模型参数烘焙

# 在 PPQ 中我们提供了数十种不同的 QuantizationOptimizationPass
# 你可以组合它们从而实现自定义的功能，也可以继承 QuantizationOptimizationPass 基类
# 从而创造出新的量化优化过程
# ------------------------------------------------------------
pipeline = PFL.Pipeline(
    [
        QuantizeSimplifyPass(),
        QuantizeFusionPass(activation_type=quantizer.activation_fusion_types),
        ParameterQuantizePass(),
        RuntimeCalibrationPass(),
        QuantAlignmentPass(force_overlap=True),
        LearnedStepSizePass(steps=1000, is_scale_trainable=True, lr=1e-5, block_size=4, collecting_device='cuda'),
        ParameterBakingPass(),
    ]
)

with ENABLE_CUDA_KERNEL():
    # 调用管线完成量化
    pipeline.optimize(
        graph=graph, dataloader=dataset, verbose=True, calib_steps=32, collate_fn=collate_fn, executor=executor
    )

    # 执行量化误差分析
    graphwise_error_analyse(graph=graph, running_device='cuda', dataloader=dataset, collate_fn=collate_fn)

# ------------------------------------------------------------
# 在最后，我们导出计算图
# 同样地，我们根据不同推理框架的需要，写了一堆不同的网络导出逻辑
# 你通过参数 platform 告诉 PPQ 你的模型最终将部署在何处，
# PPQ 则会返回一个对应的 GraphExporter 对象，它将负责将 PPQ 的量化信息
# 翻译成推理框架所需的内容。你也可以自己写一个 GraphExporter 类并注册到 PPQ 框架中来。
# ------------------------------------------------------------
exporter = PFL.Exporter(platform=TargetPlatform.TRT_INT8)
exporter.export(file_path='Quantized.onnx', graph=graph)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class AttributeI(object):
    __slots__ = ['_tab']

    @classmethod
    def SizeOf(cls):
        return 8

    # AttributeI
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # AttributeI
    def I(self): return self._tab.Get(flatbuffers.number_types.Int64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))

def CreateAttributeI(builder, i):
    builder.Prep(8, 8)
    builder.PrependInt64(i)
    return builder.Offset()


class AttributeIT(object):

    # AttributeIT
    def __init__(self):
        self.i = 0  # type: int

    @classmethod
    def InitFromBuf(cls, buf, pos):
        attributeI = AttributeI()
        attributeI.Init(buf, pos)
        return cls.InitFromObj(attributeI)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, attributeI):
        x = AttributeIT()
        x._UnPack(attributeI)
        return x

    # AttributeIT
    def _UnPack(self, attributeI):
        if attributeI is None:
            return
        self.i = attributeI.I()

    # AttributeIT
    def Pack(self, builder):
        return CreateAttributeI(builder, self.i)



import torch
import torch.utils.data
import torchvision
from absl import logging

# 装一下下面这个库
from pytorch_quantization import nn as quant_nn

logging.set_verbosity(logging.FATAL)  # Disable logging as they are too noisy in notebook

from pytorch_quantization import quant_modules

# 调用这个 quant_modules.initialize()
# 然后你正常训练就行了 ...
quant_modules.initialize()

model = torchvision.models.resnet50()
model.cuda()

# Quantization Aware Training is based on Straight Through Estimator (STE) derivative approximation.
# It is some time known as “quantization aware training”.
# We don’t use the name because it doesn’t reflect the underneath assumption.
# If anything, it makes training being “unaware” of quantization because of the STE approximation.

# After calibration is done, Quantization Aware Training is simply select a training schedule and continue training the calibrated model.
# Usually, it doesn’t need to fine tune very long. We usually use around 10% of the original training schedule,
# starting at 1% of the initial training learning rate,
# and a cosine annealing learning rate schedule that follows the decreasing half of a cosine period,
# down to 1% of the initial fine tuning learning rate (0.01% of the initial training learning rate).

# Quantization Aware Training (Essentially a discrete numerical optimization problem) is not a solved problem mathematically.
# Based on our experience, here are some recommendations:

# For STE approximation to work well, it is better to use small learning rate.
# Large learning rate is more likely to enlarge the variance introduced by STE approximation and destroy the trained network.

# Do not change quantization representation (scale) during training, at least not too frequently.
# Changing scale every step, it is effectively like changing data format (e8m7, e5m10, e3m4, et.al) every step,
# which will easily affect convergence.

# https://github.com/NVIDIA/TensorRT/blob/main/tools/pytorch-quantization/examples/finetune_quant_resnet50.ipynb


def export_onnx(model, onnx_filename, batch_onnx):
    model.eval()
    quant_nn.TensorQuantizer.use_fb_fake_quant = (
        True  # We have to shift to pytorch's fake quant ops before exporting the model to ONNX
    )
    opset_version = 13

    # Export ONNX for multiple batch sizes
    print("Creating ONNX file: " + onnx_filename)
    dummy_input = torch.randn(batch_onnx, 3, 224, 224, device='cuda')  # TODO: switch input dims by model
    torch.onnx.export(
        model,
        dummy_input,
        onnx_filename,
        verbose=False,
        opset_version=opset_version,
        enable_onnx_checker=False,
        do_constant_folding=True,
    )
    return True



# ---------------------------------------------------------------
# 这个脚本向你展示了如何使用 Onnxruntime 对 PPQ 导出的模型进行推理
# Onnxruntime 提供一系列 providers 实现不同硬件上的神经网络推理

# CPUExecutionProvider, CUDAExecutionProvider 是 Onnxruntime 官方提供的
# TensortExecutionProvider 是 Nvidia 提供的
# 不同 Provider 对模型格式有不一样的要求，PPQ 导出的是 CPUExecutionProvider 格式的模型

# Onnxruntime 没写 INT8 算子的 CUDA 实现，因此当你的模型使用 Onnxruntime 进行部署时，如果使用
# CUDAExecutionProvider, 你无需考虑量化加速
# ---------------------------------------------------------------

import torch
import torchvision

import esp_ppq
import esp_ppq.api as API

calibration_dataloader = [torch.rand(size=[1, 3, 224, 224]).cuda()]
model = torchvision.models.shufflenet_v2_x1_0().cuda()

with API.ENABLE_CUDA_KERNEL():
    quantized = API.quantize_torch_model(
        model=model,
        calib_dataloader=calibration_dataloader,
        calib_steps=8,
        input_shape=[1, 3, 224, 224],
        platform=esp_ppq.TargetPlatform.ONNXRUNTIME,
    )

API.export_ppq_graph(quantized, platform=esp_ppq.TargetPlatform.ONNXRUNTIME, graph_save_to='Quantized.onnx')

API.export_ppq_graph(quantized, platform=esp_ppq.TargetPlatform.ONNX, graph_save_to='FP32.onnx')

from esp_ppq.utils.OnnxruntimeUtil import Benchmark, Profile

Benchmark('FP32.onnx', providers=['CPUExecutionProvider'])
Benchmark('Quantized.onnx', providers=['CPUExecutionProvider'])

Profile('FP32.onnx', providers=['CPUExecutionProvider'])
Profile('Quantized.onnx', providers=['CPUExecutionProvider'])



import argparse
from typing import Iterable

import torch
import torchvision
from torch.utils.data import DataLoader
from torchvision.models.mobilenetv2 import MobileNet_V2_Weights

from esp_ppq import QuantizationSettingFactory
from esp_ppq.api import espdl_quantize_torch

BATCHSIZE = 4
INPUT_SHAPE = [3, 224, 224]
DEVICE = 'cpu'
ESPDL_MODEL_PATH = "mobilenet_v2.espdl"


class BaseInferencer:
    def __init__(self, args):
        # get quantization config.
        self.num_of_bits = args.bits
        self.target = args.target
        # Load a pretrained mobilenet v2 model
        self.model = torchvision.models.mobilenet.mobilenet_v2(weights=MobileNet_V2_Weights.IMAGENET1K_V1)
        self.model = self.model.to(DEVICE)

    def load_calibration_dataset(self) -> Iterable:
        return [torch.randn(size=INPUT_SHAPE) for _ in range(BATCHSIZE * 2)]

    def __call__(self):
        def collate_fn(batch: torch.Tensor) -> torch.Tensor:
            return batch.to(DEVICE)

        # create a setting for quantizing your network with ESPDL.
        quant_setting = QuantizationSettingFactory.espdl_setting()

        # Load training data for creating a calibration dataloader.
        calibration_dataset = self.load_calibration_dataset()
        calibration_dataloader = DataLoader(dataset=calibration_dataset, batch_size=BATCHSIZE, shuffle=False)

        # quantize your model.
        quant_ppq_graph = espdl_quantize_torch(
            model=self.model,
            espdl_export_file=ESPDL_MODEL_PATH,
            calib_dataloader=calibration_dataloader,
            calib_steps=8,
            input_shape=[1] + INPUT_SHAPE,
            target=self.target,
            num_of_bits=self.num_of_bits,
            collate_fn=collate_fn,
            setting=quant_setting,
            device=DEVICE,
            error_report=False,
            skip_export=False,
            export_test_values=True,
            verbose=1,
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="")
    parser.add_argument(
        "-t",
        "--target",
        type=str,
        default="esp32p4",
        help="esp32p4, esp32s3 or c, (defaults: esp32p4).",
    )
    parser.add_argument(
        "-b",
        "--bits",
        type=int,
        default=8,
        help="the number of bits, support 8 or 16, (defaults: 8).",
    )
    args = parser.parse_args()
    BaseInferencer(args)()



import torch
import torch.nn.functional as F

from .default import DEFAULT_BACKEND_TABLE

NXP_BACKEND_TABLE = DEFAULT_BACKEND_TABLE.copy()


def Resize_forward(op, input_value, device=None):
    """NXP Platform has a custimized resize operation implementation, which
    gives different result from torch.nn.Resize. To correctly simulate hardware
    beviour and have the same result with NXP, it is necessary to force resize
    to run with nearest mode. Any other mode of resize will be ignored by this
    function.

    Args:
        op ([type]): [description]
        input_value ([type]): [description]
        device ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """
    input_data = input_value[0]
    # Not used roi
    # roi  = input_value[1] if len(input_value) > 1 else None
    scales = input_value[2] if len(input_value) > 2 else None
    sizes = input_value[-1].tolist() if len(input_value) == 4 else None
    mode = 'nearest'

    # If 'size' is specified, then set scales to empty data (zero shape) in this operator's input list.
    if sizes is None or len(sizes) == 0:
        sizes = None
        if scales.numel() == 1:
            scales = scales.item()
        else:
            assert scales.numel() % 2 == 0
            scales = scales[-2].cpu().numpy().tolist()
    else:
        # the sizes in onnx is 4-D while in pytorch is 2-D
        # check the dim.0 & dim.1 is equal, then remain dim.2 and dim.3
        scales = None
        assert sizes[:2] == list(input_data.shape[:2])
        sizes = sizes[2:]

    trans_mode = op.attributes.get('coordinate_transformation_mode', 'half_pixel')
    if trans_mode == 'align_corners':
        output = F.interpolate(input_data, sizes, scales, mode, align_corners=True)
    else:
        output = F.interpolate(input_data, sizes, scales, mode)
    return output


# When you trying to implement a custimized function for ppl_dsp platform
# Be aware that you can just overwrite part of DEFAULT_DISPATCHING_TABLE
# rather than rewrite all dispatching table.
# here an example was given: Sample_Forward
def Sample_Forward():
    return None


NXP_BACKEND_TABLE['Sample_Function'] = Sample_Forward
# NXP_DISPATCHING_TABLE['Resize'] = Resize_forward



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Dimension(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Dimension()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsDimension(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Dimension
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Dimension
    def Value(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from FlatBuffers.Dl.DimensionValue import DimensionValue
            obj = DimensionValue()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Dimension
    def Denotation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def DimensionStart(builder):
    builder.StartObject(2)

def Start(builder):
    DimensionStart(builder)

def DimensionAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0)

def AddValue(builder, value):
    DimensionAddValue(builder, value)

def DimensionAddDenotation(builder, denotation):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(denotation), 0)

def AddDenotation(builder, denotation):
    DimensionAddDenotation(builder, denotation)

def DimensionEnd(builder):
    return builder.EndObject()

def End(builder):
    return DimensionEnd(builder)

import FlatBuffers.Dl.DimensionValue
try:
    from typing import Optional
except:
    pass

class DimensionT(object):

    # DimensionT
    def __init__(self):
        self.value = None  # type: Optional[FlatBuffers.Dl.DimensionValue.DimensionValueT]
        self.denotation = None  # type: str

    @classmethod
    def InitFromBuf(cls, buf, pos):
        dimension = Dimension()
        dimension.Init(buf, pos)
        return cls.InitFromObj(dimension)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, dimension):
        x = DimensionT()
        x._UnPack(dimension)
        return x

    # DimensionT
    def _UnPack(self, dimension):
        if dimension is None:
            return
        if dimension.Value() is not None:
            self.value = FlatBuffers.Dl.DimensionValue.DimensionValueT.InitFromObj(dimension.Value())
        self.denotation = dimension.Denotation()

    # DimensionT
    def Pack(self, builder):
        if self.value is not None:
            value = self.value.Pack(builder)
        if self.denotation is not None:
            denotation = builder.CreateString(self.denotation)
        DimensionStart(builder)
        if self.value is not None:
            DimensionAddValue(builder, value)
        if self.denotation is not None:
            DimensionAddDenotation(builder, denotation)
        dimension = DimensionEnd(builder)
        return dimension



from esp_ppq.core import QuantizationStates
from esp_ppq.IR import BaseGraph, GraphExporter
from esp_ppq.IR.quantize import QuantableOperation


class ExtensionExporter(GraphExporter):
    """ExtensionExporter is an empty exporter for you to implement customized
    logic. rewrite function export in order to dump ppq graph to disk.

    use export_ppq_graph(..., platform=TargetPlatform.EXTENSION) to invoke this class.

    Args:
        GraphExporter ([type]): [description]
    """

    def __init__(self) -> None:
        super().__init__()

    def export(self, file_path: str, graph: BaseGraph, config_path: str = None):
        """Sample Export Function -- export all quantization params into txt"""

        if config_path is None:
            raise ValueError('Can not export ppq quantization params, cause configuration path is empty.')
        with open(file=config_path, mode='w') as file:
            for op in graph.operations.values():
                if not isinstance(op, QuantableOperation):
                    continue

                for cfg, var in op.config_with_variable:
                    if QuantizationStates.can_export(cfg.state):
                        pass



#
# SPDX-FileCopyrightText: Copyright (c) 1993-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import logging
import os
import sys

import numpy as np
import pycuda.autoinit
import pycuda.driver as cuda
import tensorrt as trt

try:
    # Sometimes python does not understand FileNotFoundError
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

EXPLICIT_BATCH = 1 << (int)(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
logging.basicConfig(level=logging.INFO)
logging.getLogger("EngineBuilder").setLevel(logging.INFO)
log = logging.getLogger("EngineBuilder")


def GiB(val):
    return val * 1 << 30


def add_help(description):
    parser = argparse.ArgumentParser(description=description, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    args, _ = parser.parse_known_args()


def find_sample_data(description="Runs a TensorRT Python sample", subfolder="", find_files=[], err_msg=""):
    '''
    Parses sample arguments.

    Args:
        description (str): Description of the sample.
        subfolder (str): The subfolder containing data relevant to this sample
        find_files (str): A list of filenames to find. Each filename will be replaced with an absolute path.

    Returns:
        str: Path of data directory.
    '''

    # Standard command-line arguments for all samples.
    kDEFAULT_DATA_ROOT = os.path.join(os.sep, "usr", "src", "tensorrt", "data")
    parser = argparse.ArgumentParser(description=description, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "-d",
        "--datadir",
        help="Location of the TensorRT sample data directory, and any additional data directories.",
        action="append",
        default=[kDEFAULT_DATA_ROOT],
    )
    args, _ = parser.parse_known_args()

    def get_data_path(data_dir):
        # If the subfolder exists, append it to the path, otherwise use the provided path as-is.
        data_path = os.path.join(data_dir, subfolder)
        if not os.path.exists(data_path):
            if data_dir != kDEFAULT_DATA_ROOT:
                print("WARNING: " + data_path + " does not exist. Trying " + data_dir + " instead.")
            data_path = data_dir
        # Make sure data directory exists.
        if not (os.path.exists(data_path)) and data_dir != kDEFAULT_DATA_ROOT:
            print(
                "WARNING: {:} does not exist. Please provide the correct data path with the -d option.".format(
                    data_path
                )
            )
        return data_path

    data_paths = [get_data_path(data_dir) for data_dir in args.datadir]
    return data_paths, locate_files(data_paths, find_files, err_msg)


def locate_files(data_paths, filenames, err_msg=""):
    """
    Locates the specified files in the specified data directories.
    If a file exists in multiple data directories, the first directory is used.

    Args:
        data_paths (List[str]): The data directories.
        filename (List[str]): The names of the files to find.

    Returns:
        List[str]: The absolute paths of the files.

    Raises:
        FileNotFoundError if a file could not be located.
    """
    found_files = [None] * len(filenames)
    for data_path in data_paths:
        # Find all requested files.
        for index, (found, filename) in enumerate(zip(found_files, filenames)):
            if not found:
                file_path = os.path.abspath(os.path.join(data_path, filename))
                if os.path.exists(file_path):
                    found_files[index] = file_path

    # Check that all files were found
    for f, filename in zip(found_files, filenames):
        if not f or not os.path.exists(f):
            raise FileNotFoundError(
                "Could not find {:}. Searched in data paths: {:}\n{:}".format(filename, data_paths, err_msg)
            )
    return found_files


# Simple helper data class that's a little nicer to use than a 2-tuple.
class HostDeviceMem(object):
    def __init__(self, host_mem, device_mem):
        self.host = host_mem
        self.device = device_mem

    def __str__(self):
        return "Host:\n" + str(self.host) + "\nDevice:\n" + str(self.device)

    def __repr__(self):
        return self.__str__()


# Allocates all buffers required for an engine, i.e. host/device inputs/outputs.
def allocate_buffers(engine):
    inputs = []
    outputs = []
    bindings = []
    stream = cuda.Stream()
    for binding in engine:
        size = trt.volume(engine.get_binding_shape(binding)) * engine.max_batch_size
        dtype = trt.nptype(engine.get_binding_dtype(binding))
        # Allocate host and device buffers
        host_mem = cuda.pagelocked_empty(size, dtype)
        device_mem = cuda.mem_alloc(host_mem.nbytes)
        # Append the device buffer to device bindings.
        bindings.append(int(device_mem))
        # Append to the appropriate list.
        if engine.binding_is_input(binding):
            inputs.append(HostDeviceMem(host_mem, device_mem))
        else:
            outputs.append(HostDeviceMem(host_mem, device_mem))
    return inputs, outputs, bindings, stream


# This function is generalized for multiple inputs/outputs.
# inputs and outputs are expected to be lists of HostDeviceMem objects.
def do_inference(context, bindings, inputs, outputs, stream, batch_size=1):
    # Transfer input data to the GPU.
    [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]
    # Run inference.
    context.execute_async(batch_size=batch_size, bindings=bindings, stream_handle=stream.handle)
    # Transfer predictions back from the GPU.
    [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
    # Synchronize the stream
    stream.synchronize()
    # Return only the host outputs.
    return [out.host for out in outputs]


# This function is generalized for multiple inputs/outputs for full dimension networks.
# inputs and outputs are expected to be lists of HostDeviceMem objects.
def do_inference_v2(context, bindings, inputs, outputs, stream):
    # Transfer input data to the GPU.
    [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]
    # Run inference.
    context.execute_async_v2(bindings=bindings, stream_handle=stream.handle)
    # Transfer predictions back from the GPU.
    [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
    # Synchronize the stream
    stream.synchronize()
    # Return only the host outputs.
    return [out.host for out in outputs]


class EngineBuilder:
    """
    Parses an ONNX graph and builds a TensorRT engine from it.
    """

    def __init__(self, verbose=False):
        """
        :param verbose: If enabled, a higher verbosity level will be set on the TensorRT logger.
        """
        self.trt_logger = trt.Logger(trt.Logger.INFO)
        if verbose:
            self.trt_logger.min_severity = trt.Logger.Severity.VERBOSE

        trt.init_libnvinfer_plugins(self.trt_logger, namespace="")

        self.builder = trt.Builder(self.trt_logger)
        self.config = self.builder.create_builder_config()
        self.config.max_workspace_size = 8 * (2**30)  # 8 GB

        self.batch_size = None
        self.network = None
        self.parser = None

    def create_network(self, onnx_path):
        """
        Parse the ONNX graph and create the corresponding TensorRT network definition.
        :param onnx_path: The path to the ONNX graph to load.
        """
        network_flags = 1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)

        self.network = self.builder.create_network(network_flags)
        self.parser = trt.OnnxParser(self.network, self.trt_logger)

        onnx_path = os.path.realpath(onnx_path)
        with open(onnx_path, "rb") as f:
            if not self.parser.parse(f.read()):
                log.error("Failed to load ONNX file: {}".format(onnx_path))
                for error in range(self.parser.num_errors):
                    log.error(self.parser.get_error(error))
                sys.exit(1)

        inputs = [self.network.get_input(i) for i in range(self.network.num_inputs)]
        outputs = [self.network.get_output(i) for i in range(self.network.num_outputs)]

        log.info("Network Description")
        for input in inputs:
            self.batch_size = input.shape[0]
            log.info("Input '{}' with shape {} and dtype {}".format(input.name, input.shape, input.dtype))
        for output in outputs:
            log.info("Output '{}' with shape {} and dtype {}".format(output.name, output.shape, output.dtype))
        assert self.batch_size > 0
        self.builder.max_batch_size = self.batch_size

    def create_engine(
        self,
        engine_path,
        precision,
        calib_input=None,
        calib_cache=None,
        calib_num_images=25000,
        calib_batch_size=8,
        calib_preprocessor=None,
    ):
        """
        Build the TensorRT engine and serialize it to disk.
        :param engine_path: The path where to serialize the engine to.
        :param precision: The datatype to use for the engine, either 'fp32', 'fp16' or 'int8'.
        :param calib_input: The path to a directory holding the calibration images.
        :param calib_cache: The path where to write the calibration cache to, or if it already exists, load it from.
        :param calib_num_images: The maximum number of images to use for calibration.
        :param calib_batch_size: The batch size to use for the calibration process.
        :param calib_preprocessor: The ImageBatcher preprocessor algorithm to use.
        """
        engine_path = os.path.realpath(engine_path)
        engine_dir = os.path.dirname(engine_path)
        os.makedirs(engine_dir, exist_ok=True)
        log.info("Building {} Engine in {}".format(precision, engine_path))

        inputs = [self.network.get_input(i) for i in range(self.network.num_inputs)]

        if precision == "fp16":
            if not self.builder.platform_has_fast_fp16:
                log.warning("FP16 is not supported natively on this platform/device")
            else:
                self.config.set_flag(trt.BuilderFlag.FP16)

        with self.builder.build_engine(self.network, self.config) as engine, open(engine_path, "wb") as f:
            log.info("Serializing engine to file: {:}".format(engine_path))
            f.write(engine.serialize())



from esp_ppq.utils.TensorRTUtil import Benchmark, Profiling

print('Profiling with Int8 Model')
Profiling(engine_file='Output/INT8.engine')
print('-------------------------------------------')

print('Profiling with Fp16 Model')
Profiling(engine_file='Output/FP16.engine')
print('-------------------------------------------')

print('Profiling with Fp32 Model')
Profiling(engine_file='Output/FP32.engine')
print('-------------------------------------------')



"""
This File Contains Legacy Dispatchers.
Refer to esp_ppq.scheduler.perseus for updated implemenation.
"""

from typing import Dict, Set

from esp_ppq.core import TargetPlatform
from esp_ppq.IR import BaseGraph
from esp_ppq.IR.search import SearchableGraph

from .base import GraphDispatcher, SOI_generators, SOI_receivers, reverse_tracing_pattern, value_tracing_pattern


class AggresiveDispatcher(GraphDispatcher):
    """Graph Dispatcher cuts a graph into parts, each part of graph will
    dispatch to a specific platform for further execution and quantization.

    For the most part, all operations within graph can be partitioned into quantable operations,
        Shape-Or-Index related operations and remaining operations, all sub classes of GraphDispatcher will
        give an implementation of function "dispatch" to send all operations to their proper platform.

    ATTENTION: platform attribute will greatly affect quantizer's quantization logic, and the execution result.
        If operation is sent to a quantable platform, then its inputs and outputs will be quantized if necessary.
        if operation is classified as shape-or-index related operation, then its execution will be taken with cpu.
        if operation is sent to a fp32 platform, then its inputs and outputs shall never be quantized.

    ATTENTION: this dispatcher will insert necessary DeviceSwitch operations
        between shape-or-index operations and others.
    """

    def __init__(self, graph: BaseGraph) -> None:
        super().__init__()
        self.graph = graph

    def dispatch(
        self,
        quant_platform: TargetPlatform = TargetPlatform.UNSPECIFIED,
        fp32_platform: TargetPlatform = TargetPlatform.FP32,
        SOI_platform: TargetPlatform = TargetPlatform.SOI,
        **kwargs,
    ) -> Dict[str, TargetPlatform]:
        """Graph Dispatcher splits a graph into parts, each part of graph will
        be sent to a specific platform for further execution and quantization.

        There are 3 default platform during dispatching:
            quant_platform - all quantable parts of graph will be dispatched to this platform
            SOI_platform   - Aka. Shape or Index related operations will be dispatched to this platform.
            fp32_platform  - there are some operations receiving results from both quant_platform and SOI_platform,
                they will be dispatched to fp32_platform.

        ATTENTION: Quantization follows this dispatching,
            and only the operations within quantable platform will be quantized in the future.

        ATTENTION: this dispatcher will insert necessary DeviceSwitch operations between
            shape-or-index operations and others.

        Args:
            graph (BaseGraph): graph object which going to be dispatched by this dispatcher.

            quant_types(Set[str]): all quantable types for given platforms.

            quant_platform (TargetPlatform):
                platform object where quantable parts will goes to.

            SOI_platform (TargetPlatform):
                platform object where SOI parts will goes to.

            fp32_platform (TargetPlatform):
                platform object where remaining parts will goes to.

        Returns:
            Dict[str, TargetPlatform]: [description]
        """
        graph = self.graph
        recivers, generators = SOI_receivers(graph), SOI_generators(graph)
        search_engine, SOI_operations, FP32_operations = SearchableGraph(graph), set(recivers), set()

        quant_operations = search_engine.opset_matching(
            sp_expr=lambda x: x.is_computing_op,
            rp_expr=value_tracing_pattern,
            ep_expr=lambda x: x.type in {'Shape', 'TopK', 'NonMaxSuppression'} or x.is_boundary,
            direction='down',
        )
        # remove shape operations from computing ops.
        quant_operations.filter(lambda x: x.type == 'Shape')

        # we assume all 'Shape', 'NonMaxSuppression', 'ConstantOfShape', 'Topk' operations are SOI generators.
        shape_forward_matching = search_engine.opset_matching(
            sp_expr=lambda x: x in generators and x.type not in {'Constant'},
            rp_expr=value_tracing_pattern,
            ep_expr=lambda x: x in recivers or x in quant_operations or x.is_boundary,
            direction='down',
        )

        # update matchings, ready for further searching.
        SOI_operations.update(shape_forward_matching)
        generators.update(SOI_operations)

        while True:
            # there are some particular cases where a single matching can not handle.
            # to cover all shape-related operations, a reverse matching is required.
            shape_backward_matching = search_engine.opset_matching(
                sp_expr=lambda x: x in SOI_operations and x.type != 'Shape' and x not in quant_operations,
                rp_expr=reverse_tracing_pattern,
                ep_expr=lambda x: x in generators or x in quant_operations or x.is_boundary,
                direction='up',
            )

            if all([(op in SOI_operations) for op in shape_backward_matching]):
                break
            # update matchings
            SOI_operations.update(shape_backward_matching)

        # generate dispatching table.
        dispatching_table = {}
        for operation in graph.operations.values():
            if operation in SOI_operations:
                dispatching_table[operation.name] = SOI_platform
            elif operation in quant_operations:
                dispatching_table[operation.name] = quant_platform
            else:
                dispatching_table[operation.name] = fp32_platform

            # move Topk, Shape, NonMaxSuppression to platform as same as their input.
            if operation.type in {'Shape', 'TopK', 'NonMaxSuppression'}:
                if operation.inputs[0].source_op is not None:
                    dispatching_table[operation.name] = operation.inputs[0].source_op.platform
                else:
                    dispatching_table[operation.name] = quant_platform

            # move activations to the platform same as their input.
            if operation.is_linear_activation:
                source_op = operation.inputs[0].source_op
                if source_op is not None:
                    dispatching_table[operation.name] = dispatching_table[source_op.name]

        return dispatching_table


class ConservativeDispatcher(GraphDispatcher):
    """Graph Dispatcher cuts a graph into parts, each part of graph will
    dispatch to a specific platform for further execution and quantization.

    For the most part, all operations within graph can be partitioned into quantable operations,
        Shape-Or-Index related operations and remaining operations, all sub classes of GraphDispatcher will
        give an implementation of function "dispatch" to send all operations to their proper platform.

    Conservative Dispatcher cuts graph in a conservative way, which means it takes as much as possible operations
        into fp32 platform.

    ATTENTION: platform attribute will greatly affect quantizer's quantization logic, and the execution result.
        If operation is sent to a quantable platform, then its inputs and outputs will be quantized if necessary.
        if operation is classified as shape-or-index related operation, then its execution will be taken with cpu.
        if operation is sent to a fp32 platform, then its inputs and outputs shall never be quantized.

    ATTENTION: this dispatcher will insert necessary DeviceSwitch operations
        between shape-or-index operations and others.
    """

    def __init__(self, graph: BaseGraph) -> None:
        super().__init__()
        self.graph = graph

    def dispatch(
        self,
        quant_types: Set[str],
        quant_platform: TargetPlatform = TargetPlatform.UNSPECIFIED,
        fp32_platform: TargetPlatform = TargetPlatform.FP32,
        SOI_platform: TargetPlatform = TargetPlatform.SOI,
        **kwargs,
    ) -> Dict[str, TargetPlatform]:
        """Graph Dispatcher splits a graph into parts, each part of graph will
        be sent to a specific platform for further execution and quantization.

        There are 3 default platform during dispatching:
            quant_platform - all quantable parts of graph will be dispatched to this platform
            SOI_platform   - Aka. Shape or Index related operations will be dispatched to this platform.
            fp32_platform  - there are some operations receiving results from both quant_platform and SOI_platform,
                they will be dispatched to fp32_platform.

        ATTENTION: Quantization follows this dispatching,
            and only the operations within quantable platform will be quantized in the future.

        ATTENTION: this dispatcher will insert necessary DeviceSwitch operations between
            shape-or-index operations and others.

        Args:
            graph (BaseGraph): graph object which going to be dispatched by this dispatcher.

            quant_types(Set[str]): all quantable types for given platforms.

            quant_platform (TargetPlatform):
                platform object where quantable parts will goes to.

            fp32_platform (TargetPlatform):
                platform object where SOI parts will goes to.

            SOI_platform (TargetPlatform):
                platform object where remaining parts will goes to.

        Returns:
            Dict[str, TargetPlatform]: [description]
        """
        graph = self.graph
        recivers, generators = SOI_receivers(graph), SOI_generators(graph)
        search_engine, SOI_operations = SearchableGraph(graph), set(recivers)

        quant_operations = search_engine.opset_matching(
            sp_expr=lambda x: x.is_computing_op,
            rp_expr=value_tracing_pattern,
            ep_expr=lambda x: (x.type not in quant_types) or x.is_boundary,
            direction='down',
        )
        quant_operations.filter(lambda x: x.type not in quant_types)

        computing_extensions = search_engine.opset_matching(
            sp_expr=lambda x: x.is_computing_op,
            rp_expr=value_tracing_pattern,
            ep_expr=lambda x: x.type in {'Shape', 'TopK', 'NonMaxSuppression'} or x.is_boundary,
            direction='down',
        )

        # we assume all 'Shape', 'NonMaxSuppression', 'ConstantOfShape', 'Topk' operations are SOI generators.
        shape_forward_matching = search_engine.opset_matching(
            sp_expr=lambda x: x in generators and x.type not in {'Constant'},
            rp_expr=value_tracing_pattern,
            ep_expr=lambda x: (x in recivers or x in quant_operations or x.is_boundary or x.is_computing_op),
            direction='down',
        )

        # remove computing operations and quant operations from matching
        shape_forward_matching.filter(lambda x: x.is_computing_op or x in quant_operations)

        # update matchings, ready for further searching.
        SOI_operations.update(shape_forward_matching)

        while True:
            # there are some particular cases where a single matching can not handle.
            # to cover all shape-related operations, a reverse matching is required.
            shape_backward_matching = search_engine.opset_matching(
                sp_expr=lambda x: x in SOI_operations and x.type != 'Shape',
                rp_expr=reverse_tracing_pattern,
                ep_expr=lambda x: (x in SOI_operations or x in quant_operations or x.is_boundary or x.is_computing_op),
                direction='up',
            )

            # remove computing operations and quant operations from matching
            shape_backward_matching.filter(lambda x: x.is_computing_op or x in quant_operations)

            if all([(op in SOI_operations) for op in shape_backward_matching]):
                break

            # update matchings
            SOI_operations.update(shape_backward_matching)

        # generate dispatching table.
        dispatching_table = {}
        for operation in graph.operations.values():
            if operation in SOI_operations and operation not in computing_extensions:
                dispatching_table[operation.name] = SOI_platform
            elif operation in quant_operations:
                dispatching_table[operation.name] = quant_platform
            else:
                dispatching_table[operation.name] = fp32_platform

        for operation in graph.operations.values():
            # move Topk, Shape, NonMaxSuppression to the platform same as their input.
            if operation.type in {'Shape', 'TopK', 'NonMaxSuppression'}:
                source_op = operation.inputs[0].source_op
                if source_op is not None:
                    dispatching_table[operation.name] = dispatching_table[source_op.name]
                else:
                    dispatching_table[operation.name] = fp32_platform

            # move activations to the platform same as their input.
            if operation.is_linear_activation:
                source_op = operation.inputs[0].source_op
                if source_op is not None:
                    dispatching_table[operation.name] = dispatching_table[source_op.name]

        return dispatching_table


class PPLNNDispatcher(GraphDispatcher):
    """Graph Dispatcher cuts a graph into parts, each part of graph will
    dispatch to a specific platform for further execution and quantization.

    For the most part, all operations within graph can be partitioned into quantable operations,
        Shape-Or-Index related operations and remaining operations, all sub classes of GraphDispatcher will
        give an implementation of function "dispatch" to send all operations to their proper platform.

    Conv only Dispatcher cuts graph in a conservative way, which means it takes as much as possible operations
        into fp32 platform.

    ATTENTION: platform attribute will greatly affect quantizer's quantization logic, and the execution result.
        If operation is sent to a quantable platform, then its inputs and outputs will be quantized if necessary.
        if operation is classified as shape-or-index related operation, then its execution will be taken with cpu.
        if operation is sent to a fp32 platform, then its inputs and outputs shall never be quantized.

    ATTENTION: this dispatcher will insert necessary DeviceSwitch operations
        between shape-or-index operations and others.
    """

    def __init__(self, graph: BaseGraph) -> None:
        super().__init__()
        self.graph = graph

    def dispatch(
        self,
        quant_types: Set[str],
        quant_platform: TargetPlatform = TargetPlatform.UNSPECIFIED,
        fp32_platform: TargetPlatform = TargetPlatform.FP32,
        SOI_platform: TargetPlatform = TargetPlatform.SOI,
        **kwargs,
    ) -> Dict[str, TargetPlatform]:
        """Graph Dispatcher splits a graph into parts, each part of graph will
        be sent to a specific platform for further execution and quantization.

        There are 3 default platform during dispatching:
            quant_platform - all quantable parts of graph will be dispatched to this platform
            SOI_platform   - Aka. Shape or Index related operations will be dispatched to this platform.
            fp32_platform  - there are some operations receiving results from both quant_platform and SOI_platform,
                they will be dispatched to fp32_platform.

        ATTENTION: Quantization follows this dispatching,
            and only the operations within quantable platform will be quantized in the future.

        ATTENTION: this dispatcher will insert necessary DeviceSwitch operations between
            shape-or-index operations and others.

        Args:
            graph (BaseGraph): graph object which going to be dispatched by this dispatcher.

            quant_types(Set[str]): all quantable types for given platforms.

            quant_platform (TargetPlatform): =
                platform object where quantable parts will goes to.

            SOI_platform (TargetPlatform):
                platform object where SOI parts will goes to.

            fp32_platform (TargetPlatform):
                platform object where remaining parts will goes to.

        Returns:
            Dict[str, TargetPlatform]: [description]
        """
        graph = self.graph
        recivers, generators = SOI_receivers(graph), SOI_generators(graph)
        search_engine, SOI_operations = SearchableGraph(graph), set(recivers)

        quant_operations = search_engine.opset_matching(
            sp_expr=lambda x: x.type == 'Conv',
            rp_expr=lambda x, y: value_tracing_pattern(x, y) and y.type in quant_types,
            ep_expr=lambda x: x.type == 'Conv',
            direction='down',
        )

        computing_extensions = search_engine.opset_matching(
            sp_expr=lambda x: x.is_computing_op,
            rp_expr=value_tracing_pattern,
            ep_expr=lambda x: x.type in {'Shape', 'TopK', 'NonMaxSuppression'} or x.is_boundary,
            direction='down',
        )

        # we assume all 'Shape', 'NonMaxSuppression', 'ConstantOfShape', 'Topk' operations are SOI generators.
        shape_forward_matching = search_engine.opset_matching(
            sp_expr=lambda x: x in generators and x.type not in {'Constant'},
            rp_expr=value_tracing_pattern,
            ep_expr=lambda x: (x in recivers or x in quant_operations or x.is_boundary or x.is_computing_op),
            direction='down',
        )

        # remove computing operations and quant operations from matching
        shape_forward_matching.filter(lambda x: x.is_computing_op or x in quant_operations)

        # update matchings, ready for further searching.
        SOI_operations.update(shape_forward_matching)

        while True:
            # there are some particular cases where a single matching can not handle.
            # to cover all shape-related operations, a reverse matching is required.
            shape_backward_matching = search_engine.opset_matching(
                sp_expr=lambda x: x in SOI_operations and x.type != 'Shape',
                rp_expr=reverse_tracing_pattern,
                ep_expr=lambda x: (x in SOI_operations or x in quant_operations or x.is_boundary or x.is_computing_op),
                direction='up',
            )

            # remove computing operations and quant operations from matching
            shape_backward_matching.filter(lambda x: x.is_computing_op or x in quant_operations)

            if all([(op in SOI_operations) for op in shape_backward_matching]):
                break

            # update matchings
            SOI_operations.update(shape_backward_matching)

        # generate dispatching table.
        dispatching_table = {}
        for operation in graph.operations.values():
            if operation in SOI_operations and operation not in computing_extensions:
                dispatching_table[operation.name] = SOI_platform
            elif operation in quant_operations:
                dispatching_table[operation.name] = quant_platform
            else:
                dispatching_table[operation.name] = fp32_platform

        for operation in graph.operations.values():
            # move Topk, Shape, NonMaxSuppression to the platform same as their input.
            if operation.type in {'Shape', 'TopK', 'NonMaxSuppression'}:
                source_op = operation.inputs[0].source_op
                if source_op is not None:
                    dispatching_table[operation.name] = dispatching_table[source_op.name]
                else:
                    dispatching_table[operation.name] = fp32_platform

            # move activations to the platform same as their input.
            if operation.is_linear_activation:
                source_op = operation.inputs[0].source_op
                if source_op is not None:
                    dispatching_table[operation.name] = dispatching_table[source_op.name]

        return dispatching_table


class PointDispatcher(ConservativeDispatcher):
    """Graph Dispatcher cuts a graph into parts, each part of graph will
    dispatch to a specific platform for further execution and quantization.

    For the most part, all operations within graph can be partitioned into quantable operations,
        Shape-Or-Index related operations and remaining operations, all sub classes of GraphDispatcher will
        give an implementation of function "dispatch" to send all operations to their proper platform.

    Point Dispatch send all computing op to quantable platform, while other ops remain unchanged.

    ATTENTION: platform attribute will greatly affect quantizer's quantization logic, and the execution result.
        If operation is sent to a quantable platform, then its inputs and outputs will be quantized if necessary.
        if operation is classified as shape-or-index related operation, then its execution will be taken with cpu.
        if operation is sent to a fp32 platform, then its inputs and outputs shall never be quantized.

    ATTENTION: this dispatcher will insert necessary DeviceSwitch operations
        between shape-or-index operations and others.
    """

    def __init__(self, graph: BaseGraph) -> None:
        super().__init__()
        self.graph = graph

    def dispatch(
        self,
        quant_types: Set[str],
        quant_platform: TargetPlatform = TargetPlatform.UNSPECIFIED,
        fp32_platform: TargetPlatform = TargetPlatform.FP32,
        SOI_platform: TargetPlatform = TargetPlatform.SOI,
        **kwargs,
    ) -> Dict[str, TargetPlatform]:
        """Graph Dispatcher splits a graph into parts, each part of graph will
        be sent to a specific platform for further execution and quantization.

        There are 3 default platform during dispatching:
            quant_platform - all quantable parts of graph will be dispatched to this platform
            SOI_platform   - Aka. Shape or Index related operations will be dispatched to this platform.
            fp32_platform  - there are some operations receiving results from both quant_platform and SOI_platform,
                they will be dispatched to fp32_platform.

        ATTENTION: Quantization follows this dispatching,
            and only the operations within quantable platform will be quantized in the future.

        ATTENTION: this dispatcher will insert necessary DeviceSwitch operations between
            shape-or-index operations and others.

        Args:
            graph (BaseGraph): graph object which going to be dispatched by this dispatcher.

            quant_types(Set[str]): all quantable types for given platforms.

            quant_platform (TargetPlatform): =
                platform object where quantable parts will goes to.

            SOI_platform (TargetPlatform):
                platform object where SOI parts will goes to.

            fp32_platform (TargetPlatform):
                platform object where remaining parts will goes to.

        Returns:
            Dict[str, TargetPlatform]: [description]
        """
        graph = self.graph

        dispatch_table = ConservativeDispatcher.dispatch(
            graph=graph,
            quant_types=quant_types,
            quant_platform=quant_platform,
            fp32_platform=fp32_platform,
            SOI_platform=SOI_platform,
            kwargs=kwargs,
        )

        skip_ops = set()
        for op in graph.operations.values():
            if op in skip_ops:
                continue
            if op.type in quant_types and op.is_computing_op:
                dispatch_table[op.name] = quant_platform
            else:
                if dispatch_table[op.name] == quant_platform:
                    dispatch_table[op.name] = fp32_platform

        return dispatch_table
